__author__ = "Lukas Leufen"
__date__ = '2019-11-15'


import logging
import argparse
from typing import Union, Dict, Any

from src import helpers
from src.modules.run_environment import RunEnvironment


DEFAULT_STATIONS = ['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087', 'DEBY052', 'DEBY032', 'DEBW022', 'DEBY004',
                    'DEBY020', 'DEBW030', 'DEBW037', 'DEBW031', 'DEBW015', 'DEBW073', 'DEBY039', 'DEBW038', 'DEBW081',
                    'DEBY075', 'DEBW040', 'DEBY053', 'DEBW059', 'DEBW027', 'DEBY072', 'DEBW042', 'DEBW039', 'DEBY001',
                    'DEBY113', 'DEBY089', 'DEBW024', 'DEBW004', 'DEBY037', 'DEBW056', 'DEBW029', 'DEBY068', 'DEBW010',
                    'DEBW026', 'DEBY002', 'DEBY079', 'DEBW084', 'DEBY049', 'DEBY031', 'DEBW019', 'DEBW001', 'DEBY063',
                    'DEBY005', 'DEBW046', 'DEBW103', 'DEBW052', 'DEBW034', 'DEBY088', ]
DEFAULT_VAR_ALL_DICT = {'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                        'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                        'pblheight': 'maximum'}


class ExperimentSetup(RunEnvironment):
    """
    params:
    trainable: Train new model if true, otherwise try to load existing model
    """

    def __init__(self, parser_args=None, var_all_dict=None, stations=None, network=None, station_type=None, variables=None,
                 statistics_per_var=None, start=None, end=None, window_history=None, target_var="o3", target_dim=None,
                 window_lead_time=None, dimensions=None, interpolate_dim=None, interpolate_method=None,
                 limit_nan_fill=None, train_start=None, train_end=None, val_start=None, val_end=None, test_start=None,
                 test_end=None, use_all_stations_on_all_data_sets=True, trainable=False, fraction_of_train=None,
                 experiment_path=None):

        # create run framework
        super().__init__()

        # experiment setup
        self._set_param("data_path", helpers.prepare_host())
        self._set_param("trainable", trainable, default=False)
        self._set_param("fraction_of_training", fraction_of_train, default=0.8)

        # set experiment name
        exp_date = self._get_parser_args(parser_args).get("experiment_date")
        exp_name, exp_path = helpers.set_experiment_name(experiment_date=exp_date, experiment_path=experiment_path)
        self._set_param("experiment_name", exp_name)
        self._set_param("experiment_path", exp_path)
        helpers.check_path_and_create(self.data_store.get("experiment_path", "general"))

        # setup for data
        self._set_param("var_all_dict", var_all_dict, default=DEFAULT_VAR_ALL_DICT)
        self._set_param("stations", stations, default=DEFAULT_STATIONS)
        self._set_param("network", network, default="AIRBASE")
        self._set_param("station_type", station_type, default=None)
        self._set_param("variables", variables, default=list(self.data_store.get("var_all_dict", "general").keys()))
        self._set_param("statistics_per_var", statistics_per_var, default=self.data_store.get("var_all_dict", "general"))
        self._set_param("start", start, default="1997-01-01", scope="general")
        self._set_param("end", end, default="2017-12-31", scope="general")
        self._set_param("window_history", window_history, default=13)

        # target
        self._set_param("target_var", target_var, default="o3")
        self._set_param("target_dim", target_dim, default='variables')
        self._set_param("window_lead_time", window_lead_time, default=3)

        # interpolation
        self._set_param("dimensions", dimensions, default={'new_index': ['datetime', 'Stations']})
        self._set_param("interpolate_dim", interpolate_dim, default='datetime')
        self._set_param("interpolate_method", interpolate_method, default='linear')
        self._set_param("limit_nan_fill", limit_nan_fill, default=1)

        # train parameters
        self._set_param("start", train_start, default="1997-01-01", scope="general.train")
        self._set_param("end", train_end, default="2007-12-31", scope="general.train")

        # validation parameters
        self._set_param("start", val_start, default="2008-01-01", scope="general.val")
        self._set_param("end", val_end, default="2009-12-31", scope="general.val")

        # test parameters
        self._set_param("start", test_start, default="2010-01-01", scope="general.test")
        self._set_param("end", test_end, default="2017-12-31", scope="general.test")

        # use all stations on all data sets (train, val, test)
        self._set_param("use_all_stations_on_all_data_sets", use_all_stations_on_all_data_sets, default=True)

    def _set_param(self, param: str, value: Any, default: Any = None, scope: str = "general") -> None:
        if value is None and default is not None:
            value = default
        self.data_store.put(param, value, scope)
        logging.debug(f"set experiment attribute: {param}({scope})={value}")

    @staticmethod
    def _get_parser_args(args: Union[Dict, argparse.Namespace]) -> Dict:
        """
        Transform args to dict if given as argparse.Namespace
        :param args: either a dictionary or an argument parser instance
        :return: dictionary with all arguments
        """
        if isinstance(args, argparse.Namespace):
            return args.__dict__
        elif isinstance(args, dict):
            return args
        else:
            return {}


if __name__ == "__main__":

    formatter = '%(asctime)s - %(levelname)s: %(message)s  [%(filename)s:%(funcName)s:%(lineno)s]'
    logging.basicConfig(format=formatter, level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('--experiment_date', metavar='--exp_date', type=str, nargs=1, default=None,
                        help="set experiment date as string")
    parser_args = parser.parse_args()
    with RunEnvironment():
        setup = ExperimentSetup(parser_args, stations=['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087'])
